import logging
import os
import ssl

import paho.mqtt.client as mqtt
import msgpack

log = logging.getLogger("snap2mqtt-client")
MQTT_LOG = logging.getLogger("mqtt")


class SNAP2MQTTClientError(Exception):
    pass


def _packet_dispatch(pkt, cb):
    try:
        cb(pkt)
    except:
        log.exception("Dispatching packet callback")


def _rpc_dispatch(pkt, cb):
    try:
        cb(pkt[b'src'], *pkt[b'args'])
    except:
        log.exception("Dispatching RPC callback")


def _data_dispatch(pkt, cb):
    try:
        cb(pkt[b'src'], pkt[b'data'])
    except:
        log.exception("Dispatching Data callback")


def _multi_dispatch(pkt, cb):
    if pkt[b'type'] in [b'uRPC', b'mRPC', b'dmRPC']:
        _rpc_dispatch(pkt, cb)
    elif pkt[b'type'] in [b'uData', b'mData']:
        _data_dispatch(pkt, cb)
    else:
        log.warn("Unknown packet type: %s", pkt['type'])


class BaseClient(object):

    def __init__(self, router_address, mqtt_host='localhost', mqtt_port=1883, ssl_opts={}, dispatcher=None):
        """Create a new client, connected to the SNAP2MQTT Router on the MQTT Broker."""
        self.mqtt_client = None
        self.router_address = router_address
        self.bridge_address = None
        self.root_topic = 'snap/{}'.format(router_address)
        self.subscriptions = {}

        self._connect_callbacks = []
        self._address_change_callbacks = []
        self._disconnect_callbacks = []
        self._publish_callbacks = []

        self.mqtt_client = mqtt.Client()

        self.mqtt_client.on_connect = self._on_mqtt_connect
        self.mqtt_client.on_disconnect = self._on_mqtt_disconnect
        self.mqtt_client.on_message = self._on_mqtt_message
        self.mqtt_client.on_log = self._on_mqtt_log
        self.mqtt_client.on_subscribe = self._on_mqtt_subscribe

        if 'mqtt_ca_certs' in ssl_opts:
            # Use SSL
            ca_certs = os.path.expanduser(ssl_opts['mqtt_ca_certs'])

            self.mqtt_client.tls_set(ca_certs, tls_version=ssl.PROTOCOL_TLSv1_2)

        if 'mqtt_user' in ssl_opts:
            mqtt_user = ssl_opts['mqtt_user']
            mqtt_pass = ssl_opts.get('mqtt_pass', None)

            self.mqtt_client.username_pw_set(mqtt_user, mqtt_pass)

        self.dispatcher = dispatcher or _packet_dispatch

        self.mqtt_client.connect(mqtt_host, mqtt_port, 60)

    def on_connect(self, callback):
        if callback not in self._connect_callbacks:
            self._connect_callbacks.append(callback)

    def on_address_change(self, callback):
        if callback not in self._address_change_callbacks:
            self._address_change_callbacks.append(callback)

    def on_disconnect(self, callback):
        if callback not in self._disconnect_callbacks:
            self._disconnect_callbacks.append(callback)

    def on_publish(self, callback):
        if callback not in self._disconnect_callbacks:
            self._publish_callbacks.append(callback)

    def notify_connected(self, client, userdata, flags, rc):
        for cb in self._connect_callbacks:
            try:
                cb(client, userdata, flags, rc)
            except:
                log.exception("Notifying connected")

    def notify_address_changed(self, address):
        for cb in self._address_change_callbacks:
            try:
                cb(address)
            except:
                log.exception("Notifying on address change")

    def notify_disconnected(self, client, userdata, rc):
        for cb in self._disconnect_callbacks:
            try:
                cb(client, userdata, rc)
            except:
                log.exception("on_disconnect callback")

    def notify_published(self, client, userdata, mid):
        for cb in self._publish_callbacks:
            try:
                cb(client, userdata, mid)
            except:
                log.exception("on_publish callback")

    def start(self):
        """(Async) Start the MQTT Client loop on a background thread."""
        self.mqtt_client.loop_start()

    def stop(self):
        """(Async) Stop the MQTT Client loop thread."""
        self.mqtt_client.loop_stop()

    def loop(self):
        """(Synchronous) Enter the MQTT Client loop."""
        self.mqtt_client.loop_forever()

    def poll(self, timeout=0.05, max_packets=1):
        """(Synchronous) Poll the MQTT Client once."""
        self.mqtt_client.loop(timeout=timeout, max_packets=max_packets)

    def subscribe(self, subtopic, callback):
        """Subscribe to the specified topic with the callback.

        Multiple callbacks can be subscribed to the same topic.
        Subscriptions are maintained as sets.
        The order in which subscribed callbacks get called is not
        defined.
        """

        if not callable(callback):
            raise Exception("Callback isn't callable")

        if subtopic in self.subscriptions:
            self.subscriptions[subtopic] = set((callback,)).union(self.subscriptions[subtopic])
        else:
            # New set with one callback
            self.subscriptions[subtopic] = set((callback,))

    def unsubscribe(self, subtopic, callback):
        """Unsubscribe the callback from the topic."""

        if subtopic in self.subscriptions:
            self.subscriptions[subtopic] = set(self.subscriptions[subtopic]).difference(set((callback,)))

    def send(self, pkt):
        """Publish the packet"""
        log.info("Publishing packet: %r", pkt)
        rc, mid = self.mqtt_client.publish(self.root_topic + '/send', payload=bytearray(msgpack.dumps(pkt)))
        log.info("Publish status: %r, MsgID: %r", rc, mid)
        return (rc, mid)

    def _on_mqtt_connect(self, client, userdata, flags, rc):
        """Callback for MQTT broker connection"""
        log.info("Connected to MQTT with result code %s", rc)
        self.mqtt_client.subscribe(self.root_topic + '/recv/#')
        self.mqtt_client.subscribe(self.root_topic + '/status')

        self.notify_connected(client, userdata, flags, rc)

    def _on_mqtt_publish(self, client, userdata, mid):
        self.notify_published(client, userdata, mid)

    def _on_mqtt_disconnect(self, client, userdata, rc):
        """Callback for MQTT broker disconnection"""
        if rc == mqtt.MQTT_ERR_SUCCESS:
            log.info("MQTT disconnected successfully")
        else:
            log.error("MQTT disconnect: %s", rc)

        self.notify_disconnected(client, userdata, rc)

    def _on_mqtt_subscribe(self, client, userdata, mid, granted_qos):
        log.info("MQTT Subscribed: %r, %r", mid, granted_qos)

    def _on_mqtt_log(self, client, userdata, level, buf):
        """Map from Paho MQTT log events to Python logging."""
        # TODO: move the map elsewhere, it's not instance-specific
        if level == mqtt.MQTT_LOG_ERR:
            MQTT_LOG.error(buf)
        elif level == mqtt.MQTT_LOG_WARNING:
            MQTT_LOG.warn(buf)
        elif level == mqtt.MQTT_LOG_NOTICE:
            MQTT_LOG.info(buf)
        elif level == mqtt.MQTT_LOG_INFO:
            MQTT_LOG.info(buf)
        elif level == mqtt.MQTT_LOG_DEBUG:
            MQTT_LOG.debug(buf)

    def _on_mqtt_message(self, client, userdata, msg):
        """Callback for MQTT message received"""

        log.info("Received MQTT Message: {0.topic} == {0.payload}".format(msg))

        if mqtt.topic_matches_sub(self.root_topic + '/recv/#', msg.topic):
            try:
                pkt = msgpack.loads(msg.payload)
                log.debug("Got Packet: %s", pkt)

                for subtopic in self._get_matching_subscriptions(msg.topic):
                    for cb in self.subscriptions[subtopic]:
                        self.dispatcher(pkt, cb)
            except:
                log.exception("Couldn't understand message payload: %s", msg.payload)
        if mqtt.topic_matches_sub(self.root_topic + '/status', msg.topic):
            try:
                pkt = msgpack.loads(msg.payload)
                log.debug("Got Status Packet: %s", pkt)

                new_bridge_address = pkt['snap_bridge_address']

                if self.bridge_address != new_bridge_address:
                    self.notify_address_changed(new_bridge_address)

                self.bridge_address = new_bridge_address

            except:
                log.exception("Couldn't understand message payload: %s", msg.payload)

    def _get_matching_subscriptions(self, topic):
        return [s for s in self.subscriptions if mqtt.topic_matches_sub(self.root_topic + '/recv/' + s, topic)]


class SimpleClient(object):
    """SNAP2MQTT client with simplified interface:

    subscribe_rpc(func_name, cb)
    subscribe_data(cb)
    uRPC(dst, func, *args)
    mRPC(group, ttl, func, *args)
    dmRPC(dst, group, ttl, delay, func, *args)
    uData(dst, data)
    mData(dst, data)
    """

    def __init__(self, router_address, mqtt_host='localhost', mqtt_port=1883, ssl_opts={}):
        """Create a new simple client, connected to the SNAP2MQTT Router on the MQTT Broker."""

        self._client = BaseClient(router_address, mqtt_host, mqtt_port, ssl_opts, dispatcher=_multi_dispatch)

    def start(self):
        """(Async) Start the Client on a background thread."""
        self._client.start()

    def stop(self):
        """(Async) Stop the Client background thread."""
        self._client.stop()

    def loop(self):
        """(Synchronous) Enter the Client loop."""
        self._client.loop()

    def poll(self, timeout=0.05, max_packets=1):
        """(Synchronous) Poll the Client loop once."""
        self._client.poll(timeout, max_packets)

    def subscribe_rpc(self, func_name, callback):
        """Subscribe to func_name RPCs with the callback.

        RPCs of any type will trigger the callback.

        Callback signature should be f(source_addr, *args).
        """
        self._client.subscribe('rpc/+/' + func_name, callback)

    def unsubscribe_rpc(self, func_name, callback):
        """Unsubscribe the RPC callback"""
        self._client.unsubscribe('rpc/+/' + func_name, callback)

    def subscribe_data(self, callback):
        """Subscribe to data mode packets with the callback.

        Both PTP and Broadcast packets will trigger the callback.

        Callback signature must be f(source_addr, data).
        """
        self._client.subscribe('data/#', callback)

    def unsubscribe_data(self, callback):
        """Unsubscribe the data callback"""
        self._client.unsubscribe('data/#', callback)

    def subscribe_traceroute(self, callback):
        """Subscribe to traceroutes with the callback.

        Callback signature must be f(node_addr, rtt, hops)"""
        self._client.subscribe('traceroute/#', callback)

    def unsubscribe_traceroute(self, callback):
        """Unsusbscribe the traceroute callback"""
        self._client.unsubscribe("traceroute/#", callback)

    def uRPC(self, dst, func, *args):
        """Send a unicast RPC"""
        return self._client.send({b'type': b'uRPC',
                                  b'dst': dst,
                                  b'func': func,
                                  b'args': args})

    def mRPC(self, group, ttl, func, *args):
        """Send a multicast RPC"""
        return self._client.send({b'type': b'mRPC',
                                  b'group': group,
                                  b'ttl': ttl,
                                  b'func': func,
                                  b'args': args})

    def dmRPC(self, dst, group, ttl, delay, func, *args):
        """Send a directed multicast RPC"""
        return self._client.send({b'type': b'dmRPC',
                                  b'dst': dst,
                                  b'group': group,
                                  b'ttl': ttl,
                                  b'delay': delay,
                                  b'func': func,
                                  b'args': args})

    def uData(self, dst, data):
        """Send a unicast Data packet"""
        return self._client.send({b'type': b'uData',
                                  b'dst': dst,
                                  b'data': data})

    def mData(self, group, ttl, data):
        """Send a multicast Data packet"""
        return self._client.send({b'type': b'mData',
                                  b'group': group,
                                  b'ttl': ttl,
                                  b'data': data})

    def traceroute(self, dst):
        """Send a traceroute request"""
        return self._client.send({b'type': b'traceroute',
                                  b'dst': dst})
